@@ -7,10 +7,8 @@ from collections import defaultdict
 import pandas as pd
 import numpy as np
 
-from .utils import abs2rel_key, changes2list, DCML_REGEX, features2tpcs, fifths2iv, fifths2name, fifths2pc, fifths2rn, fifths2sd, \
-    map2elements, name2fifths, rel2abs_key, resolve_relative_keys, roman_numeral2fifths, \
-    series_is_minor, split_alternatives, transform
-from .transformations import compute_chord_tones, labels2global_tonic
+from .utils import abs2rel_key, changes2list, DCML_REGEX, rel2abs_key, resolve_relative_keys, series_is_minor, split_alternatives, transform
+from .transformations import compute_chord_tones, labels2global_tonic, transpose_chord_tones_by_localkey
 from .logger import function_logger
 
 
@@ -41,8 +39,8 @@ SM = SliceMaker()
 
 
 @function_logger
-def expand_labels(df, column='label', regex=None, cols={}, dropna=False, propagate=True, volta_structure=None,
-                  relative_to_global=False, chord_tones=True, absolute=False, all_in_c=False):
+def expand_labels(df, column='label', regex=None, rename={}, dropna=False, propagate=True, volta_structure=None,
+                  relative_to_global=False, chord_tones=True, absolute=False, all_in_c=False, skip_checks=False, ):
     """
     Split harmony labels complying with the DCML syntax into columns holding their various features
     and allows for additional computations and transformations.
@@ -61,7 +59,7 @@ def expand_labels(df, column='label', regex=None, cols={}, dropna=False, propaga
     regex : :obj:`re.Pattern`
         Compiled regular expression used to split the labels. It needs to have named groups.
         The group names are used as column names unless replaced by ``cols``.
-    cols : :obj:`dict`, optional
+    rename : :obj:`dict`, optional
         Dictionary to map the regex's group names to deviating column names of your choice.
     dropna : :obj:`bool`, optional
         Pass True if you want to drop rows where ``column`` is NaN/<NA>
@@ -98,7 +96,7 @@ def expand_labels(df, column='label', regex=None, cols={}, dropna=False, propaga
         Original DataFrame plus additional columns with split features.
     """
     assert sum((absolute, all_in_c)) < 2, "Chord tones can be either 'absolute' or 'all_in_c', not both."
-    assert len(df.index.names) == 1, f"""df has a MultiIndex of {len(df.index.names)} levels, implying that it has information 
+    assert df.index.nlevels, f"""df has a MultiIndex of {df.index.nlevels} levels, implying that it has information 
 from several pieces. Apply expand_labels() to one piece at a time."""
     df = df.copy()
     if regex is None:
@@ -112,129 +110,144 @@ from several pieces. Apply expand_labels() to one piece at a time."""
 
 
     for col in ['numeral', 'form', 'figbass', 'localkey', 'globalkey', 'phraseend']:
-        if not col in cols:
-            cols[col] = col
-    global_minor = f"{cols['globalkey']}_is_minor"
-
-    ### Check for too many immediate repetitions
-    not_nan = df[column].dropna()
-    immediate_repetitions = not_nan == not_nan.shift()
-    k = immediate_repetitions.sum()
-    if k > 0:
-        if k / len(not_nan.index) > 0.1:
-            logger.warning(
-                "DataFrame has many direct repetitions of labels. This function is written for lists of labels only which should have no immediate repetitions.")
-        else:
-            logger.debug(f"Immediate repetition of labels:\n{not_nan[immediate_repetitions]}")
-
+        if not col in rename:
+            rename[col] = col
 
+    if not skip_checks:
+        ### Check for too many immediate repetitions
+        not_nan = df[column].dropna()
+        immediate_repetitions = not_nan == not_nan.shift()
+        k = immediate_repetitions.sum()
+        if k > 0:
+            if k / len(not_nan.index) > 0.1:
+                logger.warning(
+                    "DataFrame has many direct repetitions of labels. This function is written for lists of labels only which should have no immediate repetitions.")
+            else:
+                logger.debug(f"Immediate repetition of labels:\n{not_nan[immediate_repetitions]}")
 
 
     ### Do the actual expansion
     df = split_alternatives(df, column=column, logger=logger)
-    df = split_labels(df, column=column, regex=regex, cols=cols, dropna=dropna, logger=logger)
-    df['chord_type'] = transform(df, features2type, [cols[col] for col in ['numeral', 'form', 'figbass']], logger=logger)
-    df = replace_special(df, regex=regex, merge=True, cols=cols, logger=logger)
+    df = split_labels(df, label_column=column, regex=regex, rename=rename, dropna=dropna, skip_checks=skip_checks, logger=logger)
+    df['chord_type'] = transform(df, features2type, [rename[col] for col in ['numeral', 'form', 'figbass']], logger=logger)
+    df = replace_special(df, regex=regex, merge=True, cols=rename, logger=logger)
 
-    ### Check phrase annotations
-    p_col = df[cols['phraseend']]
-    opening = p_col.fillna('').str.count('{')
-    closing = p_col.fillna('').str.count('}')
-    if opening.sum() != closing.sum():
-        o = df.loc[(opening > 0), ['mc', cols['phraseend']]]
-        c = df.loc[(closing > 0), ['mc', cols['phraseend']]]
-        compare = pd.concat([o.reset_index(drop=True), c.reset_index(drop=True)], axis=1).astype({'mc': 'Int64'})
-        logger.warning(f"Phrase beginning and endings don't match:\n{compare}")
+    if not skip_checks:
+        ### Check phrase annotations
+        p_col = df[rename['phraseend']]
+        opening = p_col.fillna('').str.count('{')
+        closing = p_col.fillna('').str.count('}')
+        if opening.sum() != closing.sum():
+            o = df.loc[(opening > 0), ['mc', rename['phraseend']]]
+            c = df.loc[(closing > 0), ['mc', rename['phraseend']]]
+            compare = pd.concat([o.reset_index(drop=True), c.reset_index(drop=True)], axis=1).astype({'mc': 'Int64'})
+            logger.warning(f"Phrase beginning and endings don't match:\n{compare}", extra={"message_id": (16, )})
 
+    key_cols = {col: rename[col] for col in ['localkey', 'globalkey']}
     if propagate:
-        key_cols = {col: cols[col] for col in ['localkey', 'globalkey']}
         try:
             df = propagate_keys(df, volta_structure=volta_structure, add_bool=True, **key_cols, logger=logger)
         except:
-            logger.error(f"propagate_keys() failed with\n{sys.exc_info()[1]}")
+            logger.error(f"propagate_keys() failed with\n{sys.exc_info()[1]}", extra={"message_id": (12, )})
         try:
-            df = propagate_pedal(df, cols=cols, logger=logger)
+            df = propagate_pedal(df, cols=rename, logger=logger)
         except:
-            logger.error(f"propagate_pedal() failed with\n{sys.exc_info()[1]}")
-
-        if chord_tones:
-            ct = compute_chord_tones(df, expand=True, cols=cols, logger=logger)
-            if relative_to_global or absolute or all_in_c:
-                transpose_by = transform(df, roman_numeral2fifths, [cols['localkey'], global_minor])
-                if absolute:
-                    transpose_by += transform(df, name2fifths, [cols['globalkey']])
-                ct = pd.DataFrame([transpose(tpcs, fifths) for tpcs, fifths in
-                                   zip(ct.itertuples(index=False, name=None), transpose_by.values)], index=ct.index,
-                                  columns=ct.columns)
-            df = pd.concat([df, ct], axis=1)
-
-        if relative_to_global:
-            labels2global_tonic(df, inplace=True, cols=cols, logger=logger)
+            logger.error(f"propagate_pedal() failed with\n{sys.exc_info()[1]}", extra={"message_id": (13, )})
     else:
         if chord_tones:
             logger.info("Chord tones cannot be calculated without propagating keys.")
         if relative_to_global:
             logger.info("Cannot transpose labels without propagating keys.")
+    not_a_chord = df.chord.isna()
+    if chord_tones:
+        key_cols_gapless = {col: (df[col].notna() | not_a_chord).all() for col in key_cols.values()}
+        if propagate or all(key_cols_gapless.values()):
+            ct = compute_chord_tones(df, expand=True, cols=rename, logger=logger)
+            df = values_into_df(df, ct)
+            if relative_to_global or absolute or all_in_c:
+                df = transpose_chord_tones_by_localkey(df, by_global=absolute)
+
+        if relative_to_global:
+            labels2global_tonic(df, inplace=True, cols=rename, logger=logger)
 
     if tmp_index:
         df.index = ix
 
     return df
 
-
-
-def transpose(e, n):
-    """ Add `n` to all elements `e` recursively.
-    """
-    return map2elements(e, lambda x: x + n)
-
-
-
+@function_logger
+def extract_features_from_labels(S, regex=None):
+    """Applies .str.extract(regex) on the Series and returns a DataFrame with all named capturing groups."""
+    if regex is None:
+        regex = DCML_REGEX
+    if regex.__class__ != re.compile('').__class__:
+        regex = re.compile(regex, re.VERBOSE)
+    features = list(regex.groupindex.keys())
+    extracted = S.str.extract(regex, expand=True)
+    return extracted[features].copy()
 
 @function_logger
-def split_labels(df, column, regex, cols={}, dropna=False, inplace=False, **kwargs):
+def split_labels(df, label_column='label', regex=None, rename={}, dropna=False, inplace=False, skip_checks=False, **kwargs):
     """ Split harmony labels complying with the DCML syntax into columns holding their various features.
 
     Parameters
     ----------
     df : :obj:`pandas.DataFrame`
         Dataframe where one column contains DCML chord labels.
-    column : :obj:`str`
+    label_column : :obj:`str`
         Name of the column that holds the harmony labels.
     regex : :obj:`re.Pattern`
         Compiled regular expression used to split the labels. It needs to have named groups.
         The group names are used as column names unless replaced by `cols`.
-    cols : :obj:`dict`
+    rename : :obj:`dict`
         Dictionary to map the regex's group names to deviating column names.
     dropna : :obj:`bool`, optional
         Pass True if you want to drop rows where ``column`` is NaN/<NA>
     inplace : :obj:`bool`, optional
         Pass True if you want to mutate ``df``.
     """
+    if regex is None:
+        regex = DCML_REGEX
     if regex.__class__ != re.compile('').__class__:
         regex = re.compile(regex, re.VERBOSE)
-    features = regex.groupindex.keys()
+
     if not inplace:
         df = df.copy()
-    if df[column].isna().any():
+    if df[label_column].isna().any():
         if dropna:
-            logger.debug(f"Removing NaN values from label column {column}...")
-            df = df[df[column].notna()]
+            logger.debug(f"Removing NaN values from label column {label_column}...")
+            df = df[df[label_column].notna()]
         else:
-            logger.debug(f"{column} contains NaN values.")
+            logger.debug(f"{label_column} contains NaN values.")
 
-    logger.debug(f"Applying RegEx to column {column}...")
-    spl = df[column].str.extract(regex, expand=True)  # .astype('string') # needs pandas 1.0
-    for feature in features:
-        name = cols[feature] if feature in cols else feature
-        df[name] = spl[feature]
-    mistakes = spl.isna().apply(all, axis=1) & df[column].notna()
-    if mistakes.any():
-        logger.warning(f"The following labels do not match the regEx:\n{df.loc[mistakes, :column].to_string()}")
+    logger.debug(f"Applying RegEx to column {label_column}...")
+    spl = extract_features_from_labels(df[label_column], regex=regex, logger=logger)
+    if len(rename) > 0:
+        spl.rename(columns=rename, inplace=True)
+    df = values_into_df(df, spl)
+    if not skip_checks:
+        syntax_errors = spl.isna().all(axis=1) & df[label_column].notna()
+        if syntax_errors.any():
+            logger.warning(f"The following labels do not match the regEx:\n{df.loc[syntax_errors, :label_column].to_string()}")
     if not inplace:
         return df
 
 
+def values_into_df(df, new_values):
+    features = list(new_values.columns)
+    update_columns = [col for col in features if col in df.columns]
+    new_columns = [col for col in features if col not in df.columns]
+    if len(update_columns) > 0:
+        df.loc[:, update_columns] = df[update_columns].fillna(new_values[update_columns])
+    if len(new_columns) > 0:
+        df = pd.concat([df, new_values[new_columns]], axis=1)
+        if len(update_columns) > 0:
+            all_other_columns = [col for col in df.columns if col not in features]
+            column_order = all_other_columns + features
+            df = df[column_order].copy()
+    return df
+
+
 @function_logger
 def features2type(numeral, form=None, figbass=None):
     """ Turns a combination of the three chord features into a chord type.
@@ -325,7 +338,7 @@ def replace_special(df, regex, merge=False, inplace=False, cols={}, special_map=
     special2label = {
         'It': 'viio6(b3)/V',
         'Ger': 'viio65(b3)/V',
-        'Fr': 'V7(b5)/V',
+        'Fr': 'V43(b5)/V',
     }
     special2label.update(special_map)
 
@@ -341,7 +354,7 @@ def replace_special(df, regex, merge=False, inplace=False, cols={}, special_map=
 
     logger.debug(f"Moving special symbols from {cols['numeral']} to {cols['special']}...")
     if not cols['special'] in df.columns:
-        df.insert(df.columns.get_loc(cols['numeral']), cols['special'], np.nan)
+        df.insert(df.columns.get_loc(cols['numeral']), cols['special'], pd.NA)
     df.loc[select_all_special, cols['special']] = df.loc[select_all_special, cols['numeral']]
 
     def repl_spec(frame, special, instead):
@@ -469,7 +482,7 @@ def propagate_keys(df, volta_structure=None, globalkey='globalkey', localkey='lo
         lm = f"{localkey}_is_minor"
         df[gm] = global_minor
         if df[localkey].str.contains('/').any():
-            lk = transform(df, resolve_relative_keys, [localkey, gm])
+            lk = transform(df, resolve_relative_keys, [localkey, gm], logger=logger)
         else:
             lk = df[localkey]
         local_minor = series_is_minor(lk)
@@ -542,8 +555,10 @@ def propagate_pedal(df, relative=True, drop_pedalend=True, cols={}):
             globalkeys = df.loc[section, cols['globalkey']].unique()
             assert len(globalkeys) == 1, "Several globalkeys appearing within the same organ point."
             global_minor = globalkeys[0].islower()
+            # if the localkey changes during the pedal point, the reference changes and the Roman numeral indicating
+            # the pedal note needs to be adapted
             key2pedal = {
-                key: ped if key == first_localkey else abs2rel_key(rel2abs_key(ped, first_localkey, global_minor), key,
+                key: ped if key == first_localkey else abs2rel_key(rel2abs_key(ped, first_localkey, global_minor, logger=logger), key,
                                                                    global_minor) for key in localkeys.unique()}
             logger.debug(
                 f"Pedal note {ped} has been transposed relative to other local keys within a global {'minor' if global_minor else 'major'} context: {key2pedal}")
@@ -565,135 +580,4 @@ def propagate_pedal(df, relative=True, drop_pedalend=True, cols={}):
 #                                           MOMENTARILY NOT IN USE:
 ########################################################################################################################
 
-def transform_columns(df, func, columns=None, param2col=None, inplace=False, **kwargs):
-    """ Wrapper function to use transform() on df[columns], leaving the other columns untouched.
 
-    Parameters
-    ----------
-    df : :obj:`pandas.DataFrame`
-        DataFrame where columns (or column combinations) work as function arguments.
-    func : :obj:`callable`
-        Function you want to apply to all elements in `columns`.
-    columns : :obj:`list`
-        Columns to which you want to apply `func`.
-    param2col : :obj:`dict` or :obj:`list`, optional
-        Mapping from parameter names of `func` to column names.
-        If you pass a list of column names, the columns' values are passed as positional arguments.
-        Pass None if you want to use all columns as positional arguments.
-    inplace : :obj:`bool`, optional
-        Pass True if you want to mutate `df` rather than getting an altered copy.
-    **kwargs: keyword arguments for transform()
-    """
-    if not inplace:
-        df = df.copy()
-
-    param_cols = []
-    if columns is None:
-        columns = df.columns
-    elif param2col is None:
-        pass
-    elif param2col.__class__ == dict:
-        param_cols = list(param2col.values())
-    else:
-        param_cols = list(param2col)
-
-    tmp_index = not df.index.is_unique
-    if tmp_index:
-        ix = df.index
-        df.reset_index(drop=True, inplace=True)
-
-    df.loc[:, columns] = transform(df[columns + param_cols], func, param2col, **kwargs)
-
-    if tmp_index:
-        df.index = ix
-
-    if not inplace:
-        return df
-
-
-
-def transform_note_columns(df, to, note_cols=['chord_tones', 'added_tones', 'bass_note', 'root'], minor_col='localkey_is_minor', inplace=False, **kwargs):
-    """
-    Turns columns with line-of-fifth tonal pitch classes into another representation.
-
-    Uses: transform_columns()
-
-    Parameters
-    ----------
-    df : :obj:`pandas.DataFrame`
-        DataFrame where columns (or column combinations) work as function arguments.
-    to : {'name', 'iv', 'pc', 'sd', 'rn'}
-        The tone representation that you want to get from the `note_cols`.
-
-        * 'name': Note names. Should only be used if the stacked fifths actually represent
-                absolute tonal pitch classes rather than intervals over the local tonic.
-                In other words, make sure to use 'name' only if 0 means C rather than I.
-        * 'iv':   Intervals such that 0 = 'P1', 1 = 'P5', 4 = 'M3', -3 = 'm3', 6 = 'A4',
-                -6 = 'D5' etc.
-        * 'pc':   (Relative) chromatic pitch class, or distance from tonic in semitones.
-        * 'sd':   Scale degrees such that 0 = '1', -1 = '4', -2 = 'b7' in major, '7' in minor etc.
-                This representation requires a boolean column `minor_col` which is
-                True in those rows where the stacks of fifths occur in a local minor
-                context and False for the others. Alternatively, if all pitches are
-                in the same mode or you simply want to express them as degrees of
-                particular mode, you can pass the boolean keyword argument `minor`.
-        * 'rn':   Roman numerals such that 0 = 'I', -2 = 'bVII' in major, 'VII' in minor etc.
-                Requires boolean 'minor' values, see 'sd'.
-
-    note_cols : :obj:`list`, optional
-        List of columns that hold integers or collections of integers that represent
-        stacks of fifth (0 = tonal center, 1 = fifth above, -1 = fourth above, etc).
-    minor_col : :obj:`str`, optional
-        If `to` is 'sd' or 'rn', specify a boolean column where the value is
-        True in those rows where the stacks of fifths occur in a local minor
-        context and False for the others.
-
-    """
-    transformations = {
-        'name': fifths2name,
-        'names': fifths2name,
-        'iv': fifths2iv,
-        'pc': fifths2pc,
-        'sd': fifths2sd,
-        'rn': fifths2rn,
-    }
-    assert to in transformations, "Parameter to needs to be one of {'name', 'iv', 'pc', 'sd', 'rn'}"
-    cols = [col for col in note_cols if col in df.columns]
-    if len(cols) < len(note_cols):
-        logger.warning(f"Columns {[[col for col in note_cols if not col in df.columns]]}")
-    param2col = None
-    if to in ['sd', 'rn']:
-        assert minor_col in df.columns or 'minor' in kwargs, f"'{to} representation requires a boolean column for the 'minor' argument, e.g. 'globalkey_is_minor'."
-        if not 'minor' in kwargs:
-            param2col = {'minor': minor_col}
-    func = transformations[to]
-    res = transform_columns(df, func, columns=note_cols, inplace=inplace, param2col=param2col, column_wise=True, **kwargs)
-    if not inplace:
-        return res
-
-
-@function_logger
-def chord2tpcs(chord, regex=None, **kwargs):
-    """
-    Split a chord label into its features and apply features2tpcs().
-
-    Uses: features2tpcs()
-
-    Parameters
-    ----------
-    chord : :obj:`str`
-        Chord label that can be split into the features ['numeral', 'form', 'figbass', 'changes', 'relativeroot'].
-    regex : :obj:`re.Pattern`, optional
-        Compiled regex with named groups for the five features. By default, the current version of the DCML harmony
-        annotation standard is used.
-    **kwargs :
-        arguments for features2tpcs (pass MC to show it in warnings!)
-    """
-    if regex is None:
-        regex = DCML_REGEX
-    chord_features = re.match(regex, chord)
-    assert chord_features is not None, f"{chord} does not match the regex."
-    chord_features = chord_features.groupdict()
-    numeral, form, figbass, changes, relativeroot = tuple(chord_features[f] for f in ('numeral', 'form', 'figbass', 'changes', 'relativeroot'))
-    return features2tpcs(numeral=numeral, form=form, figbass=figbass, changes=changes, relativeroot=relativeroot,
-                         logger=logger, **kwargs)
